/*
 * misc.h
 */

int     AssertCPU();
void    StartCounting();
void    EndCounting(unsigned *);
void    FastMemset(void *, unsigned, unsigned);
void    FastMemcpy(void *, void *, unsigned);


struct profile_struct
{
        int             enable;

        unsigned        cpu;            /* 68K time */
        unsigned        vdp;            /* VDP time */
        unsigned        render;         /* line rendering time */
        unsigned        blit;           /* blitting time */

        float           total;          /* for calculating total and % */

        float           cpu_percent;
        float           vdp_percent;
        float           render_percent;
        float           blit_percent;
};

extern struct profile_struct    profile;


static inline void StartProfile()
{
        if (!profile.enable)    /* only if allowed */
                return;
        StartCounting();        /* start the timestamp counter */                       
}

static inline void EndProfile(unsigned *element)
{
        if (!profile.enable)    /* only if allowed */
                return;
        EndCounting(element);   /* update element data */
}

static inline void TotalProfile()       /* calculates percents, resets */
{
        if (!profile.enable)    /* only if allowed */
                return;
        profile.total = profile.cpu + profile.render + profile.blit +
                        profile.vdp;
        profile.cpu_percent = (profile.cpu / profile.total) * 100;
        profile.vdp_percent = (profile.vdp / profile.total) * 100;
        profile.render_percent = (profile.render / profile.total) * 100;
        profile.blit_percent = (profile.blit / profile.total) * 100;

        /* reset */
        profile.cpu = profile.vdp = profile.render = profile.blit = 0;
}
